import numpy as np
from vg.compat import v2 as vg
from .._common.shape import columnize

__all__ = [
    "apply_transform",
    "compose_transforms",
]


def apply_transform(transform):
    """
    Wrap the given transformation matrix with a function which conveniently can
    be invoked with either points or a single point, returning the same. It
    applies the transformation to those points using homogeneous coordinates.

    Args:
        points (np.ndarray): The point `(3,)` or points `kx3` to transform.

    Return:
        func: A function which accepts an `np.ndarray` containing a point
        `(3,)` or points `kx3` to transform, and returns an `ndarray` of the
        same shape. Also accepts two kwargs. The first is `discard_z_coord`.
        When `True`, discard the z coordinate of the result. This is useful
        when applying viewport transformations. The second is
        `treat_input_as_vectors` which does not use the homogeneous coordinate,
        and therefore ignores translation.

    """
    vg.shape.check(locals(), "transform", (4, 4))

    def apply(points, discard_z_coord=False, treat_input_as_vector=False):
        points, is_columnized, maybe_decolumnize = columnize(
            points, (-1, 3), name="points"
        )

        homogenous_coordinate_value = 0 if treat_input_as_vector else 1
        padded_points = np.pad(
            points,
            ((0, 0), (0, 1)),
            mode="constant",
            constant_values=homogenous_coordinate_value,
        )
        transformed_padded_points = np.dot(transform, padded_points.T).T
        transformed_points = np.delete(transformed_padded_points, 3, axis=1)

        result = maybe_decolumnize(transformed_points)
        if discard_z_coord:
            return result[:, 0:2] if is_columnized else result[0:2]
        else:
            return result

    return apply


def compose_transforms(*transforms):
    """
    Compose the provided transformation matrices in order, returning a composite
    transformation.

    Args:
        transforms (list): One or more `4x4` transformation matrices.

    Return:
        np.ndarray: A `4x4` transformation matrix.
    """
    from functools import reduce

    for transform in transforms:
        vg.shape.check(locals(), "transform", (4, 4))

    if len(transforms) == 0:
        return np.eye(4)

    return reduce(np.dot, reversed(transforms))



# Imported and adapated from Trimesh
#
# Copyright (c) 2019 Michael Dawson-Haggerty
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# https://github.com/mikedh/trimesh/blob/510b4468d65ecb52759c9c660bf5c4791361d63f/trimesh/intersections.py#L429-L617
# https://github.com/mikedh/trimesh/blob/510b4468d65ecb52759c9c660bf5c4791361d63f/trimesh/constants.py

import logging
import numpy as np

from ..tri import FACE_DTYPE, quads_to_tris

log = logging.getLogger("polliwog.plane._trimesh_intersections")


class ToleranceMesh(object):
    """
    ToleranceMesh objects hold tolerance information about meshes.
    Parameters
    ----------------
    tol.merge : float
      When merging vertices, consider vertices closer than this
      to be the same vertex. Here we use the same value (1e-8)
      as SolidWorks uses, according to their documentation.
    """

    def __init__(self, **kwargs):
        # vertices closer than this should be merged
        self.merge = 1e-8


tol = ToleranceMesh()


def unique_bincount(values, minlength=0):
    """
    For arrays of integers find unique values using bin counting.
    Roughly 10x faster for correct input than np.unique

    Parameters
    --------------
    values : (n,) int
      Values to find unique members of
    minlength : int
      Maximum value that will occur in values (values.max())

    Returns
    ------------
    unique : (m,) int
      Unique values in original array
    inverse : (n,) int, optional
      An array such that unique[inverse] == values
    """
    values = np.asanyarray(values)
    if len(values.shape) != 1 or values.dtype.kind != "i":  # pragma: no cover
        raise ValueError("input must be 1D integers!")

    # count the number of occurrences of each value
    counts = np.bincount(values, minlength=minlength)

    # which bins are occupied at all
    # counts are integers so this works
    unique_bin = counts.astype(bool)

    # which values are unique
    # indexes correspond to original values
    unique = np.where(unique_bin)[0]

    # find the inverse to reconstruct original
    inverse = (np.cumsum(unique_bin) - 1)[values]

    return unique, inverse


def slice_faces_plane(
    vertices,
    faces,
    plane_normal,
    plane_origin,
    face_index=None,
    cached_dots=None,
    return_face_mapping=False,
):
    """
    Slice a mesh (given as a set of faces and vertices) with a plane, returning a
    new mesh (again as a set of faces and vertices) that is the
    portion of the original mesh to the positive normal side of the plane.
    Parameters
    ---------
    vertices : (n, 3) float
        Vertices of source mesh to slice
    faces : (n, 3) int
        Faces of source mesh to slice
    plane_normal : (3,) float
        Normal vector of plane to intersect with mesh
    plane_origin :  (3,) float
        Point on plane to intersect with mesh
    face_index : ((m,) int)
        Indexes of faces to slice. When no mask is provided, the
        default is to slice all faces.
    cached_dots : (n, 3) float
        If an external function has stored dot
        products pass them here to avoid recomputing
    return_face_mapping : bool
        When True, return the indices of the old faces to which the new faces
        correspond.
    Returns
    ----------
    new_vertices : (n, 3) float
        Vertices of sliced mesh
    new_faces : (n, 3) int
        Faces of sliced mesh
    """

    if len(vertices) == 0:
        empty = (vertices, faces)
        if return_face_mapping:
            return (*empty, np.arange(len(faces)))
        else:
            return empty

    # Construct a mask for the faces to slice.
    if face_index is None:
        mask = np.ones(len(faces), dtype=np.bool)
    else:
        mask = np.zeros(len(faces), dtype=np.bool)
        mask[face_index] = True

    if cached_dots is not None:  # pragma: no cover
        dots = cached_dots
    else:
        # dot product of each vertex with the plane normal indexed by face
        # so for each face the dot product of each vertex is a row
        # shape is the same as faces (n,3)
        dots = np.einsum("i,ij->j", plane_normal, (vertices - plane_origin).T)

    # Find vertex orientations w.r.t. faces for all triangles:
    #  -1 -> vertex "inside" plane (positive normal direction)
    #   0 -> vertex on plane
    #   1 -> vertex "outside" plane (negative normal direction)
    signs = np.zeros(len(vertices), dtype=np.int8)
    signs[dots < -tol.merge] = 1
    signs[dots > tol.merge] = -1
    signs = signs[faces]

    # Find all triangles that intersect this plane
    # onedge <- indices of all triangles intersecting the plane
    # inside <- indices of all triangles "inside" the plane (positive normal)
    signs_sum = signs.sum(axis=1, dtype=np.int8)
    signs_asum = np.abs(signs).sum(axis=1, dtype=np.int8)

    # Cases:
    # (0,0,0),  (-1,0,0),  (-1,-1,0), (-1,-1,-1) <- inside
    # (1,0,0),  (1,1,0),   (1,1,1)               <- outside
    # (1,0,-1), (1,-1,-1), (1,1,-1)              <- onedge
    onedge = np.logical_and(
        np.logical_and(signs_asum >= 2, np.abs(signs_sum) <= 1), mask
    )
    inside = np.logical_or((signs_sum == -signs_asum), ~mask)

    # Automatically include all faces that are "inside"
    new_faces = faces[inside]
    if return_face_mapping:
        new_face_mapping = inside.nonzero()[0]

    # Separate faces on the edge into two cases: those which will become
    # quads (two vertices inside plane) and those which will become triangles
    # (one vertex inside plane)
    onedge_quad = np.logical_and(onedge, signs_sum < 0).nonzero()[0]
    cut_faces_quad = faces[onedge_quad]
    cut_signs_quad = signs[onedge_quad]
    onedge_tri = np.logical_and(onedge, signs_sum >= 0).nonzero()[0]
    cut_faces_tri = faces[onedge_tri]
    cut_signs_tri = signs[onedge_tri]

    # If no faces to cut, the surface is not in contact with this plane.
    # Thus, return a mesh with only the inside faces
    if len(cut_faces_quad) + len(cut_faces_tri) == 0:

        if len(new_faces) == 0:
            # if no new faces at all return empty arrays
            empty = (
                np.zeros((0, 3), dtype=np.float64),
                np.zeros((0, 3), dtype=FACE_DTYPE),
            )
            if return_face_mapping:
                return (*empty, new_face_mapping)
            else:
                return empty

        # Renumber vertices, dropping any which have been orphaned.
        unique, inverse = unique_bincount(new_faces.ravel())
        final = (
            vertices[unique],
            inverse.reshape((-1, 3)),
        )
        if return_face_mapping:
            return (*final, new_face_mapping)
        else:
            return final

    # Extract the intersections of each triangle's edges with the plane
    o = vertices[faces][onedge]  # origins
    d = np.roll(o, -1, axis=1) - o  # directions
    num = (plane_origin - o).dot(plane_normal)  # compute num/denom
    denom = np.dot(d, plane_normal)
    denom[denom == 0.0] = 1e-12  # prevent division by zero
    dist = np.divide(num, denom)
    # intersection points for each segment
    int_points = np.einsum("ij,ijk->ijk", dist, d) + o

    # Initialize the array of new vertices with the current vertices
    new_vertices = vertices

    # Handle the case where a new quad is formed by the intersection
    # First, extract the intersection points belonging to a new quad
    quad_int_points = int_points[(signs_sum < 0)[onedge], :, :]
    num_quads = len(quad_int_points)
    if num_quads > 0:
        # Extract the vertex on the outside of the plane, then get the vertices
        # (in CCW order of the inside vertices)
        quad_int_inds_col = np.where(cut_signs_quad == 1)[1].reshape(-1, 1)
        quad_int_verts = cut_faces_quad[
            np.tile(np.arange(num_quads).reshape(num_quads, 1), 2),
            (np.repeat(quad_int_inds_col, 2, axis=1) + np.array([1, 2])) % 3,
        ]

        # Fill out new quad faces with the intersection points as vertices
        new_quad_faces = np.append(
            quad_int_verts,
            np.arange(len(new_vertices), len(new_vertices) + 2 * num_quads).reshape(
                num_quads, 2
            ),
            axis=1,
        )

        # Extract correct intersection points from int_points and order them in
        # the same way as they were added to faces
        new_quad_vertices = quad_int_points[
            np.tile(np.arange(num_quads).reshape(num_quads, 1), 2),
            (np.repeat(quad_int_inds_col, 2, axis=1) + np.array([2, 0])) % 3,
            :,
        ].reshape(2 * num_quads, 3)

        # Add new vertices to existing vertices, triangulate quads, and add the
        # resulting triangles to the new faces
        new_vertices = np.append(new_vertices, new_quad_vertices, axis=0)
        new_tri_faces_from_quads = quads_to_tris(new_quad_faces)
        new_faces = np.append(new_faces, new_tri_faces_from_quads, axis=0)
        if return_face_mapping:
            new_face_mapping = np.append(
                new_face_mapping,
                # Two new triangles have been added for each quad.
                np.repeat(onedge_quad, 2),
            )

    # Handle the case where a new triangle is formed by the intersection
    # First, extract the intersection points belonging to a new triangle
    tri_int_points = int_points[(signs_sum >= 0)[onedge], :, :]
    num_tris = len(tri_int_points)
    if num_tris > 0:
        # Extract the single vertex for each triangle inside the plane and get the
        # inside vertices (CCW order)
        tri_int_inds = np.where(cut_signs_tri == -1)[1]
        tri_int_verts = cut_faces_tri[range(num_tris), tri_int_inds].reshape(
            num_tris, 1
        )

        # Fill out new triangles with the intersection points as vertices
        new_tri_faces = np.append(
            tri_int_verts,
            np.arange(len(new_vertices), len(new_vertices) + 2 * num_tris).reshape(
                num_tris, 2
            ),
            axis=1,
        )

        # Extract correct intersection points and order them in the same way as
        # the vertices were added to the faces
        new_tri_vertices = tri_int_points[
            np.tile(np.arange(num_tris).reshape(num_tris, 1), 2),
            (np.repeat(tri_int_inds.reshape(-1, 1), 2, axis=1) + np.array([0, 2])) % 3,
            :,
        ].reshape(2 * num_tris, 3)

        # Append new vertices and new faces
        new_vertices = np.append(new_vertices, new_tri_vertices, axis=0)
        new_faces = np.append(new_faces, new_tri_faces, axis=0)
        if return_face_mapping:
            # One new triangle has been added for each triangle.
            new_face_mapping = np.append(new_face_mapping, onedge_tri)

    # Renumber vertices, dropping any which have been orphaned.
    unique, inverse = unique_bincount(new_faces.ravel())
    final = (
        new_vertices[unique],
        inverse.reshape((-1, 3)),
    )
    if return_face_mapping:
        return (*final, new_face_mapping)
    else:
        return final



import numpy as np
from vg.compat import v2 as vg
from ._affine_transform import (
    transform_matrix_for_non_uniform_scale,
    transform_matrix_for_rotation,
    transform_matrix_for_translation,
    transform_matrix_for_uniform_scale,
)
from ._apply import apply_transform
from ._rotation import rotation_from_up_and_look


class CompositeTransform:
    """
    Composite transform using homogeneous coordinates.

    Example:
        >>> transform = CompositeTransform()
        >>> transform.uniform_scale(10)
        >>> transform.reorient(up=[0, 1, 0], look=[-1, 0, 0])
        >>> transform.translate([0, -2.5, 0])
        >>> transformed_scan = transform(scan_v)
        >>> # ... register the scan here ...
        >>> untransformed_alignment = transform(alignment_v, reverse=True)

    See also:

        - *Computer Graphics: Principles and Practice*, Hughes, van Dam,
          McGuire, Sklar, Foley
        - http://gamedev.stackexchange.com/questions/72044/why-do-we-use-4x4-matrices-to-transform-things-in-3d
    """

    def __init__(self):
        # List of tuples, containing forward and reverse matrices.
        self.transforms = []

    def __call__(self, points, from_range=None, reverse=False, discard_z_coord=False):
        """
        Args:
            points (np.arraylike): Points to transform, as a 3xn array.
            from_range (tuple): The indices of the subset of the
                transformations to apply. e.g. `(0, 2)`, `(2, 4)`. When
                `None`, which is the default, apply them all.
            reverse (bool): When `True` applies the selected transformations
                in reverse. This has no effect on how range is interpreted,
                only whether the selected transformations apply in the forward
                or reverse mode.

        """
        return apply_transform(
            self.transform_matrix_for(from_range=from_range, reverse=reverse)
        )(points, discard_z_coord=discard_z_coord)

    def transform_matrix_for(self, from_range=None, reverse=False):
        """
        Return a 4x4 transformation matrix representation.

        range: The min and max indices of the subset of the transformations to
          apply. e.g. (0, 2), (2, 4). Inclusive of the min value, exclusive of
          the max value. The default is to apply them all.
        reverse: When `True` returns a matrix for the inverse transform.
          This has no effect on how range is interpreted, only whether the
          forward or reverse matrices are used.

        """
        from ._apply import compose_transforms

        if from_range is not None:
            start, stop = from_range
            selected_transforms = self.transforms[start:stop]
        else:
            selected_transforms = self.transforms

        if reverse:
            matrices = [inverse for _, inverse in reversed(selected_transforms)]
        else:
            matrices = [forward for forward, _ in selected_transforms]

        return compose_transforms(*matrices)

    def append_transform(self, forward, reverse=None):
        """
        Append an arbitrary transformation, defined by 4x4 forward and reverse
        matrices.

        The new transformation is added to the end. Return its index.

        """
        vg.shape.check(locals(), "forward", (4, 4))
        if reverse is None:
            reverse = np.linalg.inv(forward)
        else:
            vg.shape.check(locals(), "reverse", (4, 4))

        new_index = len(self.transforms)
        self.transforms.append((forward, reverse))
        return new_index

    def uniform_scale(self, factor, allow_flipping=False):
        """
        Scale by the given factor.

        Args:
            factor (float): The scale factor.

        See also:
            `non_uniform_scale()`
        """
        forward, inverse = transform_matrix_for_uniform_scale(
            factor, allow_flipping=allow_flipping, ret_inverse_matrix=True
        )
        return self.append_transform(forward, inverse)

    def non_uniform_scale(self, x_factor, y_factor, z_factor, allow_flipping=False):
        """
        Scale by the given factors along `x`, `y`, and `z`.

        Args:
            x_factor (float): The scale factor to be applied along the `x` axis.
            y_factor (float): The scale factor to be applied along the `y` axis.
            z_factor (float): The scale factor to be applied along the `z` axis.

        See also:
            `uniform_scale()`
        """
        forward, inverse = transform_matrix_for_non_uniform_scale(
            x_factor,
            y_factor,
            z_factor,
            allow_flipping=allow_flipping,
            ret_inverse_matrix=True,
        )
        return self.append_transform(forward, inverse)

    def convert_units(self, from_units, to_units):
        """
        Convert the mesh from one set of units to another.

        These calls are equivalent:

        >>> composite.convert_units(from_units='cm', to_units='m')
        >>> composite.uniform_scale(.01)

        Supports the length units from Ounce:
        https://github.com/lace/ounce/blob/master/ounce/core.py#L26
        """
        import ounce

        factor = ounce.factor(from_units, to_units)
        return self.uniform_scale(factor)

    def flip(self, dim):
        """
        Flip about one of the axes.

        Args:
            dim (int): The axis to flip about: 0 for `x`, 1 for `y`, 2 for `z`.
        """
        if dim not in (0, 1, 2):
            raise ValueError("Expected dim to be 0, 1, or 2")

        scale_factors = np.ones(3)
        scale_factors[dim] = -1.0
        return self.non_uniform_scale(*scale_factors, allow_flipping=True)

    def translate(self, translation):
        """
        Translate by the vector provided.

        Args:
            vector (np.arraylike): A 3x1 vector.
        """
        forward, inverse = transform_matrix_for_translation(
            translation, ret_inverse_matrix=True
        )
        return self.append_transform(forward, inverse)

    def reorient(self, up, look):
        """
        Reorient using up and look.
        """
        return self.rotate(rotation_from_up_and_look(up, look))

    def rotate(self, rotation):
        """
        Rotate by the given 3x3 rotation matrix or a Rodrigues vector.
        """
        forward, inverse = transform_matrix_for_rotation(
            rotation, ret_inverse_matrix=True
        )
        return self.append_transform(forward, inverse)



def project_relative_path(*path_components):
    import os

    return os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "..", "..", *path_components
    )


SCHEMA_PATH = project_relative_path("types", "src", "generated", "schema.json")



import numpy as np
from vg.compat import v2 as vg


def percentile(points, axis, percentile):
    """
    Given a cloud of points and an axis, find a point along that axis
    from the centroid at the given percentile.

    Args:
        points (np.arraylike): A `kx3` stack of points.
        axis (np.arraylike): A 3D vector specifying the direction of
            interest.
        percentile (float): The desired percentile.

    Returns:
        np.ndarray: A 3D point at the requested percentile.
    """
    k = vg.shape.check(locals(), "points", (-1, 3))
    if k < 1:
        raise ValueError("At least one point is needed")
    vg.shape.check(locals(), "axis", (3,))
    if vg.almost_zero(axis):
        raise ValueError("Axis must be non-zero")

    axis = vg.normalize(axis)
    coords_on_axis = points.dot(axis)
    selected_coord_on_axis = np.percentile(coords_on_axis, percentile)
    centroid = np.average(points, axis=0)
    return vg.reject(centroid, axis) + selected_coord_on_axis * axis


def extent(points, ret_indices=False):
    """
    Find the distance between the two farthest-most points.

    Args:
        points (np.arraylike): A `kx3` stack of points.
        ret_indices (bool): When `True`, return the indices along with the
            distance.

    Returns:
        object: With `ret_indices=False`, the distance; with
        `ret_indices=True` a tuple `(distance, first_index, second_index)`.

    Note:
        This is implemented using a brute-force method.
    """
    k = vg.shape.check(locals(), "points", (-1, 3))
    if k < 2:
        raise ValueError("At least two points are required")

    farthest_i = -1
    farthest_j = -1
    farthest_distance = -1
    for i, probe in enumerate(points):
        distances = vg.euclidean_distance(points, probe)
        this_farthest_j = np.argmax(distances)
        if distances[this_farthest_j] > farthest_distance:
            farthest_i = i
            farthest_j = this_farthest_j
            farthest_distance = distances[this_farthest_j]
    if ret_indices:
        return farthest_distance, farthest_i, farthest_j
    else:
        return farthest_distance



import numpy as np
from vg.compat import v2 as vg
from ..tri import FACE_DTYPE


def slice_triangles_by_plane(
    vertices,
    faces,
    plane_reference_point,
    plane_normal,
    faces_to_slice=None,
    ret_face_mapping=False,
):
    """
    Slice the given triangles by the given plane.

    - Triangles partially in front of the plane are sliced.
    - Triangles fully in front of the plane are kept as is.
    - Triangles fully behind the plane are culled.

    Returns a tuple: vertices, faces.
    """
    from ._trimesh_intersections import slice_faces_plane

    vg.shape.check(locals(), "vertices", (-1, 3))
    vg.shape.check(locals(), "faces", (-1, 3))
    vg.shape.check(locals(), "plane_reference_point", (3,))
    vg.shape.check(locals(), "plane_normal", (3,))
    if faces_to_slice is not None:
        vg.shape.check(locals(), "faces_to_slice", (-1,))
        assert faces_to_slice.dtype == np.bool

    result = slice_faces_plane(
        vertices=vertices,
        faces=faces,
        plane_normal=plane_normal,
        plane_origin=plane_reference_point,
        face_index=None if faces_to_slice is None else faces_to_slice.nonzero()[0],
        return_face_mapping=ret_face_mapping,
    )

    if ret_face_mapping:
        (
            vertices,
            faces,
            face_mapping,
        ) = result  # lgtm [py/mismatched-multiple-assignment]
        assert face_mapping.dtype == FACE_DTYPE
    else:
        vertices, faces = result
    assert vertices.dtype == np.float64
    assert faces.dtype == FACE_DTYPE

    return result



from ._segment_functions import (
    closest_point_of_line_segment,
    path_centroid,
    subdivide_segment,
    subdivide_segments,
)

__all__ = [
    "closest_point_of_line_segment",
    "path_centroid",
    "subdivide_segment",
    "subdivide_segments",
]



from . import _plane_functions
from ._plane_functions import *  # noqa: F401,F403
from ._plane_intersect import intersect_segment_with_plane
from ._slicing import slice_triangles_by_plane

__all__ = _plane_functions.__all__ + [
    "intersect_segment_with_plane",
    "slice_triangles_by_plane",
]



import numpy as np
from vg.compat import v2 as vg


def inflection_points(points, rise_axis, run_axis):
    """
    Find the list of vertices that preceed inflection points in a curve. The
    curve is differentiated with respect to the coordinate system defined by
    `rise_axis` and `run_axis`.

    Interestingly, `lambda x: 2*x + 1` should have no inflection points, but
    almost every point on the line is detected. It's because a zero or zero
    crossing in the second derivative is necessary but not sufficient to
    detect an inflection point. You also need a higher derivative of odd
    order that's non-zero. But that gets ugly to detect reliably using sparse
    finite differences. Just know that if you've got a straight line this
    method will go a bit haywire.

    rise_axis: A vector representing the vertical axis of the coordinate system.
    run_axis: A vector representing the the horiztonal axis of the coordinate system.

    returns: a list of points in space corresponding to the vertices that
    immediately preceed inflection points in the curve
    """
    vg.shape.check(locals(), "points", (-1, 3))
    vg.shape.check(locals(), "rise_axis", (3,))
    vg.shape.check(locals(), "run_axis", (3,))

    coords_on_run_axis = points.dot(run_axis)
    coords_on_rise_axis = points.dot(rise_axis)

    # Take the second order finite difference of the curve with respect to the
    # defined coordinate system
    finite_difference_1 = np.gradient(coords_on_rise_axis, coords_on_run_axis)
    finite_difference_2 = np.gradient(finite_difference_1, coords_on_run_axis)

    # Compare the product of all neighboring pairs of points in the second
    # derivative. If a pair of points has a negative product, then the second
    # derivative changes sign between those points. These are the inflection
    # points.
    is_inflection_point = np.concatenate(
        [finite_difference_2[:-1] * finite_difference_2[1:] <= 0, [False]]
    )

    return points[is_inflection_point]


def point_of_max_acceleration(points, rise_axis, run_axis, subdivide_by_length=None):
    """
    Find the point on a curve where the curve is maximally accelerating
    in the direction specified by `rise_axis`. `run_axis` is the horizontal
    axis along which slices are taken.

    Args:
        points (np.arraylike): A stack of points, as `kx3`. For best
            results, trim these to the area of interest before calling.
        rise_axis (np.arraylike): The vertical axis, as a 3D vector.
        run_axis (np.arraylike): The horizonal axis, as a 3D vector.
        subdivide_by_length (float): When provided, the maximum space
            between each point. The idea is keep the slice width small,
            however this constraint is applied in 3D space, not along
            the `run_axis`. For best results pass a value that is small
            relative to the changes in the geometry. When `None`, the
            points are used without modification.
    """
    from ..polyline._polyline_object import Polyline

    k = vg.shape.check(locals(), "points", (-1, 3))
    vg.shape.check(locals(), "rise_axis", (3,))
    vg.shape.check(locals(), "run_axis", (3,))

    if k < 2:
        raise ValueError("At least two points are required")

    if subdivide_by_length is not None:
        subdivided = Polyline(v=points, is_closed=False).subdivided_by_length(
            subdivide_by_length
        )
        points = subdivided.v

    coords_on_run_axis = points.dot(run_axis)
    coords_on_rise_axis = points.dot(rise_axis)

    finite_difference_1 = np.gradient(coords_on_rise_axis, coords_on_run_axis)
    finite_difference_2 = np.gradient(finite_difference_1, coords_on_run_axis)

    # `np.argmax(finite_difference_2)` produces false positives where the first
    # derivative of the next point is positive. Exclude these bogus points.
    valid_points = np.logical_and(
        np.roll(finite_difference_1, 1) > 0,
        np.roll(finite_difference_1, -1) > 0,
    )
    valid_points[0] = False
    valid_points[-1] = False

    try:
        index = np.argmax(finite_difference_2[valid_points])
    except ValueError:
        return None

    return points[valid_points][index]



import numpy as np
from vg.compat import v2 as vg


def write_canvas_points_to_png(canvas_coords, width, height, out_path):
    import png

    vg.shape.check(locals(), "canvas_coords", (-1, 2))

    bitmap = np.zeros((height, width), dtype=np.int8)
    for x, y in canvas_coords:
        x = int(x)
        y = int(y)
        if x >= 0 and x < width and y >= 0 and y < height:
            bitmap[y][x] = 255

    png.from_array(bitmap, mode="L").save(out_path)



import numpy as np
from vg.compat import v2 as vg

__all__ = [
    "transform_matrix_for_non_uniform_scale",
    "transform_matrix_for_rotation",
    "transform_matrix_for_translation",
    "transform_matrix_for_uniform_scale",
]


def _convert_33_to_44(matrix):
    """
    Transform from:
        array([[1., 2., 3.],
               [2., 3., 4.],
               [5., 6., 7.]])
    to:
        array([[1., 2., 3., 0.],
               [2., 3., 4., 0.],
               [5., 6., 7., 0.],
               [0., 0., 0., 1.]])

    """
    vg.shape.check(locals(), "matrix", (3, 3))
    result = np.pad(matrix, ((0, 1), (0, 1)), mode="constant")
    result[3][3] = 1
    return result


def transform_matrix_for_rotation(rotation, ret_inverse_matrix=False):
    """
    Create a transformation matrix from the given 3x3 rotation matrix or a
    Rodrigues vector.

    With `ret_inverse_matrix=True`, also returns a matrix which provides
    the reverse transform.
    """
    from ._rodrigues import rodrigues_vector_to_rotation_matrix

    if rotation.shape == (3, 3):
        forward3 = rotation
    else:
        vg.shape.check(locals(), "rotation", (3,))
        forward3 = rodrigues_vector_to_rotation_matrix(rotation)

    forward = _convert_33_to_44(forward3)

    if not ret_inverse_matrix:
        return forward

    # The inverse of a rotation matrix is its transpose.
    inverse = forward.T
    return forward, inverse


def transform_matrix_for_translation(translation, ret_inverse_matrix=False):
    """
    Create a transformation matrix which translates by the provided
    displacement vector.

    Forward:

        [[  1,  0,  0,  v_0 ],
        [  0,  1,  0,  v_1 ],
        [  0,  0,  1,  v_2 ],
        [  0,  0,  0,  1   ]]

    Reverse:

        [[  1,  0,  0,  -v_0 ],
        [  0,  1,  0,  -v_1 ],
        [  0,  0,  1,  -v_2 ],
        [  0,  0,  0,  1    ]]

    Args:
        vector (np.arraylike): A 3x1 vector.
    """
    vg.shape.check(locals(), "translation", (3,))

    forward = np.eye(4)
    forward[:, -1][:-1] = translation

    if not ret_inverse_matrix:
        return forward

    inverse = np.eye(4)
    inverse[:, -1][:-1] = -translation
    return forward, inverse


def transform_matrix_for_non_uniform_scale(
    x_factor, y_factor, z_factor, allow_flipping=False, ret_inverse_matrix=False
):
    """
    Create a transformation matrix that scales by the given factors along
    `x`, `y`, and `z`.

    Forward:
        [[  s_0, 0,   0,   0 ],
        [  0,   s_1, 0,   0 ],
        [  0,   0,   s_2, 0 ],
        [  0,   0,   0,   1 ]]

    Reverse:
        [[  1/s_0, 0,     0,     0 ],
        [  0,     1/s_1, 0,     0 ],
        [  0,     0,     1/s_2, 0 ],
        [  0,     0,     0,     1 ]]

    Args:
        x_factor (float): The scale factor to be applied along the `x` axis,
            which should be positive.
        y_factor (float): The scale factor to be applied along the `y` axis,
            which should be positive.
        z_factor (float): The scale factor to be applied along the `z` axis,
            which should be positive.
        allow_flipping (bool): When `True`, allows scale factors to be
            positive or negative, though not zero.
        ret_inverse_matrix (bool): When `True`, also returns a matrix which
            provides the inverse transform.
    """
    if x_factor == 0 or y_factor == 0 or z_factor == 0:
        raise ValueError("Scale factors should be nonzero")
    if not allow_flipping and (x_factor < 0 or y_factor < 0 or z_factor < 0):
        raise ValueError("Scale factors should be greater than zero")
    scale = np.array([x_factor, y_factor, z_factor])

    forward = _convert_33_to_44(np.diag(scale))

    if not ret_inverse_matrix:
        return forward

    inverse = _convert_33_to_44(np.diag(1.0 / scale))
    return forward, inverse


def transform_matrix_for_uniform_scale(
    scale_factor, allow_flipping=False, ret_inverse_matrix=False
):
    """
    Create a transformation matrix that scales by the given factor.

    Forward:
        [[  s_0, 0,   0,   0 ],
        [  0,   s_1, 0,   0 ],
        [  0,   0,   s_2, 0 ],
        [  0,   0,   0,   1 ]]

    Reverse:
        [[  1/s_0, 0,     0,     0 ],
        [  0,     1/s_1, 0,     0 ],
        [  0,     0,     1/s_2, 0 ],
        [  0,     0,     0,     1 ]]

    Args:
        factor (float): The scale factor.
        ret_inverse_matrix (bool): When `True`, also returns a matrix which
            provides the inverse transform.
    """
    if scale_factor == 0:
        raise ValueError("Scale factor should be nonzero")
    if not allow_flipping and scale_factor < 0:
        raise ValueError("Scale factor should be greater than zero")
    return transform_matrix_for_non_uniform_scale(
        scale_factor,
        scale_factor,
        scale_factor,
        allow_flipping=allow_flipping,
        ret_inverse_matrix=ret_inverse_matrix,
    )



def try_load_jsonschema_and_simplejson():
    try:
        import jsonschema
        import simplejson
    except ImportError:  # pragma: no cover
        raise ImportError(
            "To deserialize objects, install polliwog with the serialization extra: "
            + "`pip install entente[serialization]`"
        )
    return jsonschema, simplejson


def validator_for(schema_path, ref):
    jsonschema, json = try_load_jsonschema_and_simplejson()

    with open(schema_path, "r") as f:
        schema = json.load(f)

    resolver = jsonschema.RefResolver.from_schema(schema)
    return jsonschema.Draft7Validator({"$ref": ref}, resolver=resolver)



import numpy as np
from vg.compat import v2 as vg
from ._plane_functions import (
    mirror_point_across_plane,
    plane_normal_from_points,
    project_point_to_plane,
    signed_distance_to_plane,
)


class Plane:
    """
    An immutable 2D plane (not a hyperplane) in 3D space.

    Args:
        reference_point (np.ndarray): A reference point on the plane, as a
            NumPy array with three coordinates.
        normal (np.ndarray): The plane normal vector, with unit length, as a
            NumPy array with three coordinates.
        direction_decimals (int): The desired number of decimal places for
            validating that `normal` has unit length. The default is
            `DEFAULT_DIRECTION_DECIMALS`.

    Note:
        To construct a plane from a non-unit length normal, use
        `Plane.from_point_and_normal()`.
    """

    POSITION_DTYPE = np.float64
    DEFAULT_POSITION_DECIMALS = 6
    DEFAULT_DIRECTION_DECIMALS = 6

    def __init__(self, reference_point, normal, direction_decimals=None):
        if direction_decimals is None:
            direction_decimals = self.DEFAULT_DIRECTION_DECIMALS

        vg.shape.check(locals(), "reference_point", (3,))
        vg.shape.check(locals(), "normal", (3,))

        if not vg.almost_unit_length(normal, atol=0.1**direction_decimals):
            raise ValueError("normal should have unit length")

        self.reference_point = np.copy(reference_point)
        self.reference_point.setflags(write=False)

        self.normal = np.copy(normal)
        self.normal.setflags(write=False)

    def __repr__(self):
        return "<Plane of {} through {}>".format(self.normal, self.reference_point)

    @classmethod
    def from_point_and_normal(cls, reference_point, normal, direction_decimals=None):
        """
        Create a plane using the given reference point and normal vector, which
        will be normalized for you.

        Args:
            reference_point (np.ndarray): A reference point on the plane, as a
                NumPy array with three coordinates.
            normal (np.ndarray): The plane normal vector, with unit length, as a
                NumPy array with three coordinates.
            direction_decimals (int): The desired number of decimal places for
                validating that `normal` has unit length. The default is
                `DEFAULT_DIRECTION_DECIMALS`.

        Returns:
            Plane: The requested plane.
        """
        return cls(
            reference_point=reference_point,
            normal=vg.normalize(normal),
            direction_decimals=direction_decimals,
        )

    @classmethod
    def from_points(cls, p1, p2, p3):
        """
        If the points are oriented in a counterclockwise direction, the plane's
        normal extends towards you.

        """
        vg.shape.check(locals(), "p1", (3,))
        vg.shape.check(locals(), "p2", (3,))
        vg.shape.check(locals(), "p3", (3,))
        points = np.array([p1, p2, p3])
        return cls(reference_point=p1, normal=plane_normal_from_points(points))

    @classmethod
    def from_points_and_vector(cls, p1, p2, vector, direction_decimals=None):
        """
        Compute a plane which contains two given points and the given
        vector. Its reference point will be p1.

        For example, to find the vertical plane that passes through
        two landmarks:

            from_points_and_normal(p1, p2, vector)

        Another way to think about this: identify the plane to which
        your result plane should be perpendicular, and specify vector
        as its normal vector.

        Args:
            direction_decimals (int): The desired number of decimal places for
                validating that `normal` has unit length. The default is
                `DEFAULT_DIRECTION_DECIMALS`.

        """
        vg.shape.check(locals(), "p1", (3,))
        vg.shape.check(locals(), "p2", (3,))
        vg.shape.check(locals(), "vector", (3,))

        return cls.from_point_and_normal(
            reference_point=p1,
            normal=np.cross(p2 - p1, vector),
            direction_decimals=direction_decimals,
        )

    @classmethod
    def fit_from_points(cls, points):
        """
        Fits a plane whose normal is orthgonal to the first two principal axes
        of variation in the data and centered on their centroid.
        """
        vg.shape.check(locals(), "points", (-1, 3))

        eigval, eigvec = np.linalg.eig(np.cov(points.T))
        ordering = np.argsort(eigval)[::-1]
        normal = np.cross(eigvec[:, ordering[0]], eigvec[:, ordering[1]])

        centroid = points.mean(axis=0)

        return cls(centroid, normal)

    def rounded(self, position_decimals=None, direction_decimals=None):
        """
        Return a copy of this plane, with the reference point and normal rounded
        to the specified precision.

        Args:
            position_decimals (int): The desired number of decimal places for
                the reference point. The default is `DEFAULT_POSITION_DECIMALS`.
            direction_decimals (int): The desired number of decimal places for
                the normal. The default is `DEFAULT_DIRECTION_DECIMALS`.

        Returns:
            Plane: The rounded plane.
        """
        if position_decimals is None:
            position_decimals = self.DEFAULT_POSITION_DECIMALS
        if direction_decimals is None:
            direction_decimals = self.DEFAULT_DIRECTION_DECIMALS
        return Plane(
            reference_point=np.around(self.reference_point, position_decimals),
            normal=np.around(self.normal, direction_decimals),
        )

    def serialize(self, position_decimals=None, direction_decimals=None):
        """
        Return a JSON representation of this plane, with the reference point and
        normal rounded to the specified precision.

        The schema is defined in `types/src/schema.json`.

        Args:
            position_decimals (int): The desired number of decimal places for
                the reference point. The default is `DEFAULT_POSITION_DECIMALS`.
            direction_decimals (int): The desired number of decimal places for
                the normal. The default is `DEFAULT_DIRECTION_DECIMALS`.

        Returns:
            dict: The JSON representation.
        """
        rounded = self.rounded(
            position_decimals=position_decimals, direction_decimals=direction_decimals
        )
        return {
            "referencePoint": rounded.reference_point.tolist(),
            "unitNormal": rounded.normal.tolist(),
        }

    @classmethod
    def validate(cls, data):
        """
        Validate a plane JSON representation.

        The schema is defined in `types/src/schema.json`.

        Args:
            data (dict): The JSON representation.
        """
        from .._common.pathlib import SCHEMA_PATH
        from .._common.serialization import validator_for

        try:
            validator = cls._validator
        except AttributeError:
            validator = None

        if validator is None:
            validator = cls._validator = validator_for(
                schema_path=SCHEMA_PATH,
                ref="#/definitions/Plane",
            )

        validator.validate(data)

    @classmethod
    def deserialize(cls, data):
        """
        Create a Plane from the given JSON representation.

        The schema is defined in `types/src/schema.json`.

        Args:
            data (dict): The JSON representation.

        Returns:
            Plane: The deserialized plane.
        """
        cls.validate(data)

        return cls(
            reference_point=np.array(data["referencePoint"]),
            normal=np.array(data["unitNormal"]),
        )

    @property
    def equation(self):
        """
        Returns parameters `A`, `B`, `C`, `D` as a 1x4 `np.array`, where

            `Ax + By + Cz + D = 0`

        defines the plane.
        """
        A, B, C = self.normal
        D = -self.reference_point.dot(self.normal)

        return np.array([A, B, C, D])

    @property
    def canonical_point(self):
        """
        A canonical point on the plane, the one at which the normal
        would intersect the plane if drawn from the origin (0, 0, 0).

        This is computed by projecting the reference point onto the
        normal.

        This is useful for partitioning the space between two planes,
        as we do when searching for planar cross sections.

        """
        return self.reference_point.dot(self.normal) * self.normal

    def flipped(self):
        """
        Creates a new Plane with an inverted orientation.
        """
        return Plane(reference_point=self.reference_point, normal=-self.normal)

    def flipped_if(self, condition):
        """
        Conditionally flip the plane, returning `self` or a new Plane with an
        inverted orientation.
        """
        return self.flipped() if condition else self

    def sign(self, points):
        """
        Given an array of points, return an array with +1 for points in front
        of the plane (in the direction of the normal), -1 for points behind
        the plane (away from the normal), and 0 for points on the plane.

        """
        return np.sign(self.signed_distance(points))

    def points_in_front(self, points, inverted=False, ret_indices=False):
        """
        Given an array of points, return the points which lie in the
        half-space in front of it (i.e. in the direction of the plane
        normal).

        Args:
            points (np.arraylikw): An array of points.
            inverted (bool): When `True`, return the points which lie on or
                behind the plane instead.
            ret_indices (bool): When `True`, return the indices instead of the
                points themselves.

        Note:
            Use `points_on_or_in_front()` for points which lie either on the
            plane or in front of it.
        """
        sign = self.sign(points)

        if inverted:
            mask = np.less(sign, 0)
        else:
            mask = np.greater(sign, 0)

        indices = np.flatnonzero(mask)

        return indices if ret_indices else points[indices]

    def points_on_or_in_front(self, points, inverted=False, ret_indices=False):
        """
        Given an array of points, return the points which lie either on the
        plane or in the half-space in front of it (i.e. in the direction of
        the plane normal).

        Args:
            points (np.arraylikw): An array of points.
            inverted (bool): When `True`, return the points behind the plane
                instead.
            ret_indices (bool): When `True`, return the indices instead of the
                points themselves.

        Note:
            Use `points_in_front()` to get points which lie only in front of
            the plane.
        """
        sign = self.sign(points)

        if inverted:
            mask = np.less_equal(sign, 0)
        else:
            mask = np.greater_equal(sign, 0)

        indices = np.flatnonzero(mask)

        return indices if ret_indices else points[indices]

    def signed_distance(self, points):
        """
        Returns the signed distances to the given points or the signed
        distance to a single point.

        Args:
            points (np.arraylike): A 3D point or a `kx3` stack of points.

        Returns:
            depends:

            - Given a single 3D point, the distance as a NumPy scalar.
            - Given a `kx3` stack of points, an `k` array of distances.
        """
        return signed_distance_to_plane(points, self.equation)

    def distance(self, points):
        return np.absolute(self.signed_distance(points))

    def project_point(self, points):
        """
        Project a given point (or stack of points) to the plane.
        """
        return project_point_to_plane(points, self.equation)

    def mirror_point(self, points):
        """
        Mirror a point (or stack of points) to the opposite side of the plane.
        """
        return mirror_point_across_plane(points, self.equation)

    def line_xsection(self, pt, ray):
        vg.shape.check(locals(), "pt", (3,))
        vg.shape.check(locals(), "ray", (3,))
        return self._line_xsection(np.asarray(pt).ravel(), np.asarray(ray).ravel())

    def _line_xsection(self, pt, ray):
        denom = np.dot(ray, self.normal)
        if denom == 0:
            return None  # parallel, either coplanar or non-intersecting
        p = np.dot(self.reference_point - pt, self.normal) / denom
        return p * ray + pt

    def line_segment_xsection(self, a, b):
        vg.shape.check(locals(), "a", (3,))
        vg.shape.check(locals(), "b", (3,))
        return self._line_segment_xsection(np.asarray(a).ravel(), np.asarray(b).ravel())

    def _line_segment_xsection(self, a, b):
        pt = self._line_xsection(a, b - a)
        if pt is not None:
            if any(np.logical_and(pt > a, pt > b)) or any(
                np.logical_and(pt < a, pt < b)
            ):
                return None
        return pt

    def line_xsections(self, pts, rays):
        k = vg.shape.check(locals(), "pts", (-1, 3))
        vg.shape.check(locals(), "rays", (k, 3))
        denoms = np.dot(rays, self.normal)
        denom_is_zero = denoms == 0
        denoms[denom_is_zero] = np.nan
        p = np.dot(self.reference_point - pts, self.normal) / denoms
        return np.vstack([p, p, p]).T * rays + pts, ~denom_is_zero

    def line_segment_xsections(self, a, b):
        pts, pt_is_valid = self.line_xsections(a, b - a)
        pt_is_out_of_bounds = np.logical_or(
            np.any(
                np.logical_and(
                    pts[pt_is_valid] > a[pt_is_valid], pts[pt_is_valid] > b[pt_is_valid]
                ),
                axis=1,
            ),
            np.any(
                np.logical_and(
                    pts[pt_is_valid] < a[pt_is_valid], pts[pt_is_valid] < b[pt_is_valid]
                ),
                axis=1,
            ),
        )
        pt_is_valid[pt_is_valid] = ~pt_is_out_of_bounds
        pts[~pt_is_valid] = np.nan
        return pts, pt_is_valid

    def tilted(self, new_point, coplanar_point):
        """
        Create a new plane, tilted so it passes through `new_point`. Also
        specify a `coplanar_point` which the old and new planes should have
        in common.

        Args:
            new_point (np.arraylike): A point on the desired plane, with shape
                `(3,)`.
            coplanar_point (np.arraylike): The `(3,)` point which the old and
                new planes have in common.

        Returns:
            Plane: The adjusted plane.
        """
        vg.shape.check(locals(), "new_point", (3,))
        vg.shape.check(locals(), "coplanar_point", (3,))

        vector_along_old_plane = self.project_point(new_point) - coplanar_point
        vector_along_new_plane = new_point - coplanar_point
        axis_of_rotation = vg.perpendicular(vector_along_old_plane, self.normal)
        angle_between_vectors = vg.signed_angle(
            vector_along_old_plane,
            vector_along_new_plane,
            look=axis_of_rotation,
            units="rad",
        )
        new_normal = vg.rotate(
            self.normal,
            around_axis=axis_of_rotation,
            angle=angle_between_vectors,
            units="rad",
        )
        return Plane(reference_point=coplanar_point, normal=new_normal)


Plane.xy = Plane(reference_point=np.zeros(3), normal=vg.basis.z)
Plane.xy.__doc__ = "The `xy`-plane."
Plane.xz = Plane(reference_point=np.zeros(3), normal=vg.basis.y)
Plane.xz.__doc__ = "The `xz`-plane."
Plane.yz = Plane(reference_point=np.zeros(3), normal=vg.basis.x)
Plane.yz.__doc__ = "The `yz`-plane."



import numpy as np
from vg.compat import v2 as vg
from ._affine_transform import (
    transform_matrix_for_rotation,
    transform_matrix_for_translation,
)
from ._apply import compose_transforms


def world_to_view(position, target, up=vg.basis.y, inverse=False):
    """
    Create a transform matrix which sends world-space coordinates to
    view-space coordinates.

    Args:
        position (np.ndarray): The camera's position in world coordinates.
        target (np.ndarray): The camera's target in world coordinates.
            `target - position` is the "look at" vector.
        up (np.ndarray): The approximate up direction, in world coordinates.
        inverse (bool): When `True`, return the inverse transform instead.

    Returns:
        np.ndarray: The `4x4` transformation matrix, which can be used with
        `polliwog.transform.apply_transform()`.

    See also:
        https://cseweb.ucsd.edu/classes/wi18/cse167-a/lec4.pdf
        http://www.songho.ca/opengl/gl_camera.html
    """
    vg.shape.check(locals(), "position", (3,))
    vg.shape.check(locals(), "target", (3,))

    look = vg.normalize(target - position)
    left = vg.normalize(vg.cross(look, up))
    recomputed_up = vg.cross(left, look)

    rotation = transform_matrix_for_rotation(np.array([left, recomputed_up, look]))
    if inverse:
        inverse_rotation = rotation.T
        inverse_translation = transform_matrix_for_translation(position)
        return compose_transforms(inverse_rotation, inverse_translation)
    else:
        translation = transform_matrix_for_translation(-position)
        return compose_transforms(translation, rotation)


def view_to_orthographic_projection(width, height, near=0.1, far=2000, inverse=False):
    """
    Create an orthographic projection matrix with the given parameters, which
    maps points from world space to coordinates in the normalized view volume.
    These coordinates range from -1 to 1 in x, y, and z with `(-1, -1, -1)`
    at the bottom-left of the near clipping plane, and `(1, 1, 1)` at the
    top-right of the far clipping plane.

    Args:
        width (float): Width of the window, in pixels. (FIXME: Is this really
            correct?)
        height (float): Height of the window, in pixels. (FIXME: Is this really
            correct?)
        near (float): Near clipping plane. (FIXME: Clarify!)
        far (float): Far clipping plane. (FIXME: Clarify!)
        inverse (bool): When `True`, return the inverse transform instead.

    Returns:
        np.ndarray: The `4x4` transformation matrix, which can be used with
        `polliwog.transform.apply_transform()`.

    See also:
        https://cseweb.ucsd.edu/classes/wi18/cse167-a/lec4.pdf
        http://www.songho.ca/opengl/gl_projectionmatrix.html
        http://glasnost.itcarlow.ie/~powerk/GeneralGraphicsNotes/projection/orthographicprojection.html
    """
    if inverse:
        inverse_translate = np.array(
            [
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 1, (far + near) / (far - near)],
                [0, 0, 0, 1],
            ]
        )
        inverse_scale = np.array(
            [
                [width / 2, 0, 0, 0],
                [0, height / 2, 0, 0],
                [0, 0, (far - near) / -2, 0],
                [0, 0, 0, 1],
            ]
        )
        return compose_transforms(inverse_translate, inverse_scale)
    else:
        scale = np.array(
            [
                [2 / width, 0, 0, 0],
                [0, 2 / height, 0, 0],
                [0, 0, -2 / (far - near), 0],
                [0, 0, 0, 1],
            ]
        )
        translate = np.array(
            [
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 1, -(far + near) / (far - near)],
                [0, 0, 0, 1],
            ]
        )
        return compose_transforms(scale, translate)


def viewport_transform(x_right, y_bottom, x_left=0, y_top=0, inverse=False):
    """
    Create a matrix which transforms from the normalized view volume to
    screen coordinates, with a depth value ranging from 0 in front to 1 in
    back.

    No clipping is performed.

    Args:
        x_right (int): The `x` coordinate of the right of the viewport.
            (usually the width).
        y_bottom (int): The `y` coordinate of the bottom of the viewport
            (usually the height).
        x_left (int): The `x` coordinate of the left of the viewport
            (usually zero).
        y_top (int): The `y` coordinate of the top of the viewport
            (usually zero).
        inverse (bool): When `True`, return the inverse transform instead.

    Returns:
        np.ndarray: The `4x4` transformation matrix, which can be used with
        `polliwog.transform.apply_transform()`.

    See also:
        https://cseweb.ucsd.edu/classes/wi18/cse167-a/lec4.pdf
        http://glasnost.itcarlow.ie/~powerk/GeneralGraphicsNotes/projection/viewport_transformation.html
    """
    if inverse:
        inverse_translate = np.array(
            [
                [1, 0, 0, -0.5 * (x_right + x_left)],
                [0, 1, 0, -0.5 * (y_top + y_bottom)],
                [0, 0, 1, -0.5],
                [0, 0, 0, 1],
            ]
        )
        inverse_scale = np.array(
            [
                [2 / (x_right - x_left), 0, 0, 0],
                [0, 2 / (y_top - y_bottom), 0, 0],
                [0, 0, 2, 0],
                [0, 0, 0, 1],
            ]
        )
        return compose_transforms(inverse_translate, inverse_scale)
    else:
        scale = np.array(
            [
                [0.5 * (x_right - x_left), 0, 0, 0],
                [0, 0.5 * (y_top - y_bottom), 0, 0],
                [0, 0, 0.5, 0],
                [0, 0, 0, 1],
            ]
        )
        translate = np.array(
            [
                [1, 0, 0, 0.5 * (x_right + x_left)],
                [0, 1, 0, 0.5 * (y_top + y_bottom)],
                [0, 0, 1, 0.5],
                [0, 0, 0, 1],
            ]
        )
        return compose_transforms(scale, translate)


def world_to_canvas_orthographic_projection(
    width, height, position, target, zoom=1, inverse=False
):
    """
    Create a transformation matrix which composes camera, orthographic
    projection, and viewport transformations into a single operation.

    Args:
        width (float): Width of the window, in pixels. (FIXME: Is this really
            correct?)
        height (float): Height of the window, in pixels. (FIXME: Is this really
            correct?)
        position (np.ndarray): The camera's position in world coordinates.
        target (np.ndarray): The camera's target in world coordinates.
            `target - position` is the "look at" vector.
        inverse (bool): When `True`, return the inverse transform instead.

    Returns:
        np.ndarray: The `4x4` transformation matrix, which can be used with
        `polliwog.transform.apply_transform()`.

    """
    from ._apply import compose_transforms

    transforms = [
        world_to_view(position=position, target=target, inverse=inverse),
        view_to_orthographic_projection(
            width=width / zoom, height=height / zoom, inverse=inverse
        ),
        viewport_transform(x_right=width, y_bottom=height, inverse=inverse),
    ]
    if inverse:
        transforms.reverse()
    return compose_transforms(*transforms)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['polliwog',
 'polliwog._common',
 'polliwog.box',
 'polliwog.line',
 'polliwog.plane',
 'polliwog.pointcloud',
 'polliwog.polyline',
 'polliwog.segment',
 'polliwog.shapes',
 'polliwog.transform',
 'polliwog.tri']

package_data = \
{'': ['*']}

install_requires = \
['numpy<1.19.0', 'ounce>=1.1.0,<2.0', 'vg>=2.0.0']

extras_require = \
{'serialization': ['jsonschema>=4,<5', 'simplejson>=3,<4']}

setup_kwargs = {
    'name': 'polliwog',
    'version': '3.0.0a1',
    'description': '2D and 3D computational geometry library',
    'long_description': None,
    'author': 'Paul Melnikow',
    'author_email': 'github@paulmelnikow.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://polliwog.dev/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



import numpy as np

EDGE_DTYPE = np.int64


def edges_for(num_v, is_closed):
    """
    Compute edges of a polyline, as an array of vertex indices for each edge.

    Args:
        num_v (int): The number of vertices.
        is_closed (bool): Whether or not the polyline is closed.

    Returns:
        np.ndarray: The `kx2` edges array.
    """
    num_e = num_v if is_closed else num_v - 1

    if num_e == 0:
        return np.zeros((0, 2), dtype=EDGE_DTYPE)

    edges = np.vstack(
        [np.arange(num_e, dtype=EDGE_DTYPE), np.arange(num_e, dtype=EDGE_DTYPE) + 1]
    ).T
    if is_closed:
        edges[-1][1] = 0

    return edges



import numpy as np
from vg.compat import v2 as vg
from .._common.shape import check_shape_any, columnize
from ..line._line_functions import coplanar_points_are_on_same_side_of_line

FACE_DTYPE = np.int64

__all__ = [
    "FACE_DTYPE",
    "edges_of_faces",
    "surface_normals",
    "surface_area",
    "tri_contains_coplanar_point",
    "barycentric_coordinates_of_points",
    "sample",
]


def edges_of_faces(faces, normalize=True):
    """
    Given a stack of triangles expressed as vertex indices, return a
    normalized array of edges. When `normalize=True`, sort the edges so they
    more easily can be compared.
    """
    vg.shape.check(locals(), "faces", (-1, 3))
    assert faces.dtype == FACE_DTYPE

    # TODO: It's probably possible to accomplish this more efficiently. Maybe
    # with `np.pick()`?
    interleaved_edges = np.stack(
        [faces[:, 0:2], faces[:, 1:3], np.roll(faces, 1, axis=1)[:, 0:2]]
    )
    flattened_edges = np.swapaxes(interleaved_edges, 0, 1).reshape(-1, 2)
    return np.sort(flattened_edges, axis=1) if normalize else flattened_edges


def surface_normals(points, normalize=True):
    """
    Compute the surface normal of a triangle. The direction of the normal
    follows conventional counter-clockwise winding and the right-hand
    rule.

    Also works on stacked inputs (i.e. many sets of three points).
    """
    points, _, transform_result = columnize(points, (-1, 3, 3), name="points")

    p1s = points[:, 0]
    p2s = points[:, 1]
    p3s = points[:, 2]
    v1s = p2s - p1s
    v2s = p3s - p1s
    normals = vg.cross(v1s, v2s)

    if normalize:
        normals = vg.normalize(normals)

    return transform_result(normals)


def surface_area(vertices_of_tris):
    """
    Compute the surface area of a triangle or triangles.

    Args:
        vertices_of_tris (np.arraylike): A set of triangle vertices as `3x3` or
        `kx3x3`.

    Returns:
        object: For `3x3` input, a float containing the area. For `kx3x3` input,
        an array containing the area of each triangle.
    """
    points, _, transform_result = columnize(
        vertices_of_tris, (-1, 3, 3), name="vertices_of_tris"
    )

    e1s = points[:, 1] - points[:, 0]
    e2s = points[:, 2] - points[:, 0]

    cross_products = np.array(
        [
            e1s[:, 1] * e2s[:, 2] - e1s[:, 2] * e2s[:, 1],
            e1s[:, 2] * e2s[:, 0] - e1s[:, 0] * e2s[:, 2],
            e1s[:, 0] * e2s[:, 1] - e1s[:, 1] * e2s[:, 0],
        ]
    ).T
    areas = 0.5 * np.sqrt((cross_products**2).sum(axis=1))

    return transform_result(areas)


def tri_contains_coplanar_point(a, b, c, point):
    """
    Assuming `point` is coplanar with the triangle `ABC`, check if it lies
    inside it.
    """
    check_shape_any(a, (3,), (-1, 3), name="a")
    vg.shape.check(locals(), "b", a.shape)
    vg.shape.check(locals(), "c", a.shape)
    vg.shape.check(locals(), "point", a.shape)

    # Uses "same-side technique" from http://blackpawn.com/texts/pointinpoly/default.html
    return np.logical_and(
        np.logical_and(
            coplanar_points_are_on_same_side_of_line(b, c, point, a),
            coplanar_points_are_on_same_side_of_line(a, c, point, b),
        ),
        coplanar_points_are_on_same_side_of_line(a, b, point, c),
    )


def barycentric_coordinates_of_points(vertices_of_tris, points):
    """
    Compute barycentric coordinates for the projection of a set of points to a
    given set of triangles specfied by their vertices.

    These barycentric coordinates can refer to points outside the triangle.
    This happens when one of the coordinates is negative. However they can't
    specify points outside the triangle's plane. (That requires tetrahedral
    coordinates.)

    The returned coordinates supply a linear combination which, applied to the
    vertices, returns the projection of the original point the plane of the
    triangle.

    Args:
        vertices_of_tris (np.arraylike): A set of triangle vertices as `kx3x3`.
        points (np.arraylike): Coordinates of points as `kx3`.

    Returns:
        np.ndarray: Barycentric coordinates as `kx3`

    See Also:
        - https://en.wikipedia.org/wiki/Barycentric_coordinate_system
        - Heidrich, "Computing the Barycentric Coordinates of a Projected
          Point," JGT 05 (http://www.cs.ubc.ca/~heidrich/Papers/JGT.05.pdf)
    """
    k = vg.shape.check(locals(), "vertices_of_tris", (-1, 3, 3))
    vg.shape.check(locals(), "points", (k, 3))

    p = points.T
    q = vertices_of_tris[:, 0].T
    u = (vertices_of_tris[:, 1] - vertices_of_tris[:, 0]).T
    v = (vertices_of_tris[:, 2] - vertices_of_tris[:, 0]).T

    n = np.cross(u, v, axis=0)
    s = np.sum(n * n, axis=0)

    # If the triangle edges are collinear, cross-product is zero,
    # which makes "s" 0, which gives us divide by zero. So we
    # make the arbitrary choice to set s to epsv (=numpy.spacing(1)),
    # the closest thing to zero
    s[s == 0] = np.spacing(1)

    oneOver4ASquared = 1.0 / s
    w = p - q
    b2 = np.sum(np.cross(u, w, axis=0) * n, axis=0) * oneOver4ASquared
    b1 = np.sum(np.cross(w, v, axis=0) * n, axis=0) * oneOver4ASquared
    b = np.vstack((1 - b1 - b2, b1, b2))

    return b.T


RANDOM_SEED = 1337


def sample(
    vertices_of_tris,
    num_samples,
    rng=None,
    weights=None,
    ret_points=True,
    ret_face_indices=False,
):
    """
    Generate points sampled across the surface of the triangles provided.

    By default, triangles are weighted by area, and the random number seed is
    fixed, making this function deterministic.

    Args:
        vertices_of_tris (np.arraylike): A set of triangle vertices as `kx3x3`.
        num_samples (int): The number of samples desired.
        rng (np.random.Generator): A NumPy random number generator. To obtain
            random results for each invocation, pass `np.random.default_rng()`.
        weights (np.arraylike): `kx3` weights. By default, triangles are
            weighted by area.
        ret_face_indices (bool): When True, return the face indices from
            which each point was taken.

    Returns:
        object: The sampledpoints, or a tuple containing the sampled points and
        their corresponding face indices.
    """
    k = vg.shape.check(locals(), "vertices_of_tris", (-1, 3, 3))
    if not isinstance(num_samples, int):
        raise ValueError("Expected num_samples to be an int")
    if rng is None:
        rng = np.random.default_rng(RANDOM_SEED)
    elif not isinstance(rng, np.random.Generator):
        raise ValueError("Expected rng to be an instance of np.random.Generator")
    if weights is None:
        weights = surface_area(vertices_of_tris)
    else:
        vg.shape.check(locals(), "weights", (k,))

    if k == 0:
        empty_samples = np.zeros((0, 3), dtype=np.float64)
        if ret_face_indices:
            empty_face_indices = np.zeros((0))
            return empty_samples, empty_face_indices
        else:
            return empty_samples

    # Adapted from Trimesh, Copyright (c) 2019 Michael Dawson-Haggerty
    # https://github.com/mikedh/trimesh/blob/00ffb55b150d37988f6113a758c50f4092a49bd1/trimesh/sample.py
    # and
    # https://blogs.sas.com/content/iml/2020/10/19/random-points-in-triangle.html

    # Pick which triangles will be sampled.
    cumulative_weights = np.cumsum(weights)
    total_weight = cumulative_weights[-1]
    face_indices = np.searchsorted(
        cumulative_weights, rng.random(num_samples) * total_weight
    )

    v0s = vertices_of_tris[face_indices, 0]
    edge_vectors = vertices_of_tris[face_indices, 1:] - np.tile(v0s, (1, 2)).reshape(
        (-1, 2, 3)
    )

    # Generate scalar coefficients from 0-1 for each edge vector. When
    # multiplied out, this produces samples of parallelograms, of which the
    # triangle makes up half.  When the sum of the coefficients `u` and `v` is >
    # 1, this point will be outside the triangle. Transform `u` to `1 - u` and
    # `v` to `1 - v` to reflect the sample across the last edge, back onto the
    # triangle.
    coeffs = rng.random((num_samples, 2, 1))
    coeffs_needing_reflection = coeffs.sum(axis=1).ravel() > 1
    coeffs[coeffs_needing_reflection] = 1 - coeffs[coeffs_needing_reflection]

    # Multiply out the vectors to produce the samples.
    samples = v0s + (coeffs * edge_vectors).sum(axis=1)

    return (samples, face_indices) if ret_face_indices else samples
